#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Get access to the API of your homeserver."""

from __future__ import annotations

import json
import sys

from logging import critical
from logging import debug
from typing import Any
from typing import Dict
from typing import Tuple
from urllib.parse import urlparse

import requests

from matrixctl import __version__
from matrixctl.errors import InternalResponseError


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def check_type(var: Any, var_name: str, var_type: Any = str) -> None:
    """Check the type of an variable in an simplified and unified way."""
    if not isinstance(var, var_type):
        raise TypeError(
            f"{var_name} needs to be of type {var_type}, you have entered "
            f"{type(var)}."
        )


class UrlBuilder:

    """Build the URL for an API request."""

    __slots__ = (
        "__scheme",
        "__subdomain",
        "__domain",
        "__api_path",
        "__api_version",
        "__path",
    )

    def __init__(self, default_domain: str) -> None:
        self.__scheme: str = "https"
        self.__subdomain: str = "matrix"
        self.__domain: str = default_domain
        self.__api_path: str = "_synapse/admin"
        self.__api_version: str = "v2"
        self.__path: str = ""

    def _scheme(self, scheme: str) -> None:
        """Set the scheme.

        Parameters
        ----------
        scheme : str
            The scheme.

        Returns
        -------
        None

        """
        check_type(scheme, "scheme")
        self.__scheme = scheme

    def _domain(self, domain: str) -> None:
        """Set the domain.

        Parameters
        ----------
        domain : str
            The domain.

        Returns
        -------
        None

        """
        check_type(domain, "domain")
        self.__domain = domain

    def _subdomain(self, subdomain: str) -> None:
        """Set the sub-domain.

        Parameters
        ----------
        subdomain : str
            The sub-domain.

        Returns
        -------
        None

        """
        check_type(subdomain, "subdomain")
        self.__subdomain = subdomain

    def _api_path(self, api_path: str) -> None:
        """Set the API path.

        Parameters
        ----------
        api_path : str
            The API path.

        Returns
        -------
        None

        """
        check_type(api_path, "api_path")
        self.__api_path = api_path

    def _api_version(self, api_version: str) -> None:
        """Set the API version.

        Parameters
        ----------
        api_version : str
            The API version.

        Returns
        -------
        None

        """
        check_type(api_version, "api_version")
        self.__api_version = api_version

    def _path(self, path: str) -> None:
        """Set the path.

        Parameters
        ----------
        path : str
            The path.

        Returns
        -------
        None

        """
        check_type(path, "path")
        self.__path = path

    scheme = property(fset=_scheme)
    domain = property(fset=_domain)
    subdomain = property(fset=_subdomain)
    api_path = property(fset=_api_path)
    api_version = property(fset=_api_version)
    path = property(fset=_path)

    def build(self) -> str:
        """Build the URL.

        Parameters
        ----------
        None

        Returns
        -------
        url : str
            The URL.

        """
        # Url Generation
        url: str = (
            f"{self.__scheme}://"
            f"{self.__subdomain}{'.' if self.__subdomain else ''}"
            f"{self.__domain}"
            f"/{self.__api_path}"
            f"/{self.__api_version}"
            f"/{self.__path}"
        )
        debug(f"url (unparsed): {url}")
        url = urlparse(url).geturl()
        debug(f"url   (parsed): {url}")

        return url


class API:

    """Handle the REST API connection of synapse."""

    __slots__ = (
        "token",
        "domain",
        "json_format",
        "url",
        "__success_codes",
        "session",
        "__request",
        "__method",
        "__params",
        "__headers",
    )

    # (*range(200, 208), 226)
    RESPONSE_OK: Tuple[int, ...] = (
        200,
        201,
        202,
        203,
        204,
        205,
        206,
        207,
        226,
    )

    def __init__(
        self, domain: str, token: str, json_format: bool = True
    ) -> None:
        """Initialize the Api class and checks, if the parameter are there.

        :param api_domain:  The API domain (e.g. "domain.tld")
        :param api_token:   The access token of an admin
        :return:            ``None``
        """

        check_type(domain, "domain")
        check_type(token, "token")
        check_type(json_format, "json_format", bool)

        self.token: str = token
        self.json_format: bool = json_format

        self.url: UrlBuilder = UrlBuilder(domain)
        self.__success_codes: Tuple[int, ...] = self.__class__.RESPONSE_OK

        self.session = requests.Session()

        self.__method: str = "GET"
        self.__params: Dict[str, str] = {}
        self.__headers: Dict[str, str] = {
            "User-Agent": f"matrixctl{__version__}",
            "Authorization": f"Bearer {self.token}",
        }

    def _method(self, method: str) -> None:
        """Set the request method.

        Parameters
        ----------
        method : str
            Set the method. Possible methods are:
            - ``"GET"``
            - ``"POST"``
            - ``"PUT"``
            - ``"DELETE"``

        Returns
        -------
        None

        """
        method = method.upper()
        if method not in {"GET", "POST", "PUT", "DELETE"}:
            raise ValueError(
                'method needs to be one of {"GET", "POST", "PUT", "DELETE"}, '
                f"you have entered {type(method)}."
            )
        self.__method = method

    def _params(self, params: dict[str, str]) -> None:
        """Set the params.

        Parameters
        ----------
        params : dict [str, str]
            Set the params with a dict.

        Returns
        -------
        None

        """
        check_type(params, "params", dict)
        self.__params.update(params)

    def _headers(self, headers: dict[str, str]) -> None:
        """Set the headers.

        Parameters
        ----------
        headers : dict [str, str]
            Set the headers with a dict.

        Returns
        -------
        None

        """
        check_type(headers, "headers", dict)

        if self.json_format:
            self.__headers["Content-Type"] = "application/json"

        self.__headers.update(headers)

    def _success_codes(self, codes: tuple[int]) -> None:
        """Set the success code(s).

        Parameters
        ----------
        codes : tuple of str
            Set the success code(s) as tuple of strings.

        Returns
        -------
        None

        """
        check_type(codes, "codes", dict)
        self.__success_codes = codes

    method = property(fset=_method)
    params = property(fset=_params)
    headers = property(fset=_headers)
    success_codes = property(fset=_success_codes)

    def request(
        self, data: bytes | str | None | dict[str, Any] = None
    ) -> requests.Response:
        """Send an request to the synapse API and receive a response.

        Parameters
        ----------
        data : bytes or str or dict [str, Any], optional, default=None
            The payload of the request.

        Returns
        -------
        response : requests.Response
            Returns the response

        """
        url: str = self.url.build()

        if self.json_format and data is not None:
            data = json.dumps(data)

        debug_headers = self.__headers.copy()
        # len("Bearer ") = 7
        debug_headers[
            "Authorization"
        ] = f"HIDDEN (Length={len(self.__headers['Authorization']) - 7})"
        debug(f"Method: {self.__method}")
        debug(f"Headers: {debug_headers}")
        debug(f"Params: {self.__params}")
        if not isinstance(data, bytes):  # TODO: Bytes debug
            debug(f"Data: {data}")

        response = self.session.request(
            method=self.__method,
            data=data,
            url=url,
            params=self.__params,
            headers=self.__headers,
            allow_redirects=False,
        )

        if response.status_code == 302:
            critical(
                "The api request resulted in an redirect (302). "
                "This indicates, that the API might have changed, or your "
                "playbook is misconfigured.\n"
                "Please make sure your installation of matrixctl is "
                "up-to-date and your vars.yml contains:\n\n"
                "matrix_nginx_proxy_proxy_matrix_client_redirect_root_uri_to"
                '_domain: ""'
            )
            sys.exit(1)
        if response.status_code == 404:
            critical(
                "You need to make sure, that your vars.yml contains the "
                "following excessive long line:\n\n"
                "matrix_nginx_proxy_proxy_matrix_client_api_forwarded_"
                "location_synapse_admin_api_enabled: true"
            )
            sys.exit(1)

        debug(f"{response.json()=}")

        debug(f"{response.status_code=}")
        if response.status_code not in self.__success_codes:
            try:
                if response.json()["errcode"] == "M_UNKNOWN_TOKEN":
                    critical(
                        "The server rejected your access-token. "
                        "Please make sure, your access-token is correct "
                        "and up-to-date. Your access-token will change every "
                        "time, you log out."
                    )
                    sys.exit(1)
            except Exception:  # pylint: disable=broad-except
                pass

            raise InternalResponseError(payload=response)

        return response


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``serve-notice`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import error

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_server_notice(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl server-notice`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser(
        "server-notice", help="Send a server notice"
    )
    parser.add_argument(
        "username",
        help=(
            "The user which will receive the server-notice."
            "(e.g. for '@Michael@MichaelSasser.org' use 'michael')"
        ),
    )
    parser.add_argument("message", help="The message")
    parser.set_defaults(func=server_notice)


def server_notice(arg: Namespace) -> int:
    """Send a server notice to a matrix instance.

    Notes
    -----
    - It uses the synapse admin API.
    - Note that "server notices" must be enabled in homeserver.yaml before
      this API can be used.

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``.

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    toml: TOML = TOML()
    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))
    request = {
        "user_id": (f"@{arg.username}:" f"{toml.get('API', 'Domain')}"),
        "content": {
            "msgtype": "m.text",
            "body": arg.message,
        },
    }

    try:
        api.url.path = "send_server_notice"
        api.url.api_version = "v1"
        api.method = "POST"
        api.request(request)
    except InternalResponseError:
        error("The server notice was not sent.")

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use the functions of this module as printing helpers."""

from __future__ import annotations

from typing import Any


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


# TODO: Check if used and for what; type?; docs.
def human_readable_bool(b: Any) -> str:
    """Use this helper function to get a "yes" or "no" string from a "bool".

    Parameters
    ----------
    b : any
        The value to "convert".

    Returns
    -------
    answer : str
        ``"Yes"`` if expression is ``True``, or
        ``"False"`` if expression is ``False``.

    """
    if isinstance(b, str):
        b = int(b)

    if isinstance(b, int):
        b = bool(b)

    return "Yes" if b else "No"


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``adduser`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import error

from .errors import InternalResponseError
from .handlers.ansible import ansible_run
from .handlers.api import API
from .handlers.toml import TOML
from .password_helpers import ask_password
from .password_helpers import ask_question
from .password_helpers import gen_password


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_adduser(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl adduser`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by
        ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser(
        "adduser", help="Add a new matrix user"
    )
    parser.add_argument("user", help="The Username of the new user")
    parser.add_argument(
        "-p",
        "--passwd",
        help="The password of the new user. (If you don't enter a password, "
        "you will be asked later.)",
    )
    parser.add_argument(
        "-a", "--admin", action="store_true", help="Create as admin user"
    )
    parser.add_argument(
        "--ansible", action="store_true", help="Use ansible insted of the api"
    )
    parser.set_defaults(func=adduser)


def adduser(arg: Namespace) -> int:
    """Add a User to the synapse instance.

    It runs ``ask_password()`` first. If ``ask_password()`` returns ``None``
    it generates a password with ``gen_password()``. Then it gives the user
    a overview of the username, password and if the new user should be
    generated as admin (if you added the ``--admin`` argument). Next, it asks
    a question, if the entered values are correct with the ``ask_question``
    function.

    If the ``ask_question`` function returns True, it continues. If not, it
    starts from the beginning.

    Depending on the ``--ansible`` switch it runs the ``adduser`` command
    via ansible or the API

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """

    toml: TOML = TOML()
    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))

    while True:
        passwd_generated: bool = False

        if arg.passwd is None:
            arg.passwd = ask_password()

        if arg.passwd == "":
            arg.passwd = gen_password()
            passwd_generated = True

        print(f"Username: {arg.user}")

        if passwd_generated:
            print(f"Password (generated): {arg.passwd}")
        else:
            print("Password: **HIDDEN**")
        print(f"Admin:    {'yes' if arg.admin else 'no'}")

        answer = ask_question()

        if answer:
            break
        arg.passwd = None

    if arg.ansible:
        ansible_run(
            playbook=toml.get("ANSIBLE", "Playbook"),
            tags="register-user",
            extra_vars={
                "username": arg.user,
                "password": arg.passwd,
                "admin": "yes" if arg.admin else "no",
            },
        )
    else:
        try:
            api.url.path = f"users/@{arg.user}:{toml.get('API','Domain')}"
            api.method = "PUT"
            api.request({"password": arg.passwd, "admin": arg.admin})
        except InternalResponseError:
            error("The User was not added.")

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``deploy`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import debug

from .handlers.ansible import ansible_run
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_deploy(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl deploy`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by
        ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser(
        "deploy", help="Provision and deploy"
    )
    parser.set_defaults(func=deploy)


def deploy(_: Namespace) -> int:
    """Deploy the ansible playbook.

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    debug("deploy")
    toml: TOML = TOML()
    ansible_run(
        playbook=toml.get("ANSIBLE", "Playbook"),
        tags="setup-all",
    )

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``version`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import debug
from logging import error
from logging import fatal

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML
from .typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_version(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl version`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser(
        "version", help="Get the version of the Synapse instance"
    )
    parser.set_defaults(func=version)


def version(_: Namespace) -> int:
    """Get the version of the Synapse instance.

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``.
        (Unused in this function)

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    toml: TOML = TOML()
    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))
    api.url.path = "server_version"
    api.url.api_version = "v1"
    try:
        response: JsonDict = api.request().json()
    except InternalResponseError:
        fatal("Could not get the server sersion.")

        return 1
    debug(f"{response=}")
    try:
        print(f"Server Version: {response['server_version']}")
        print(f"Python Version: {response['python_version']}")
    except KeyError:
        error("MatrixCtl was not able to read the server version.")

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# flake8: noqa
# pylint: disable:undefined-variable

"""Use MatrixCtl to control, manage, provision and deploy your homeserver. """

from __future__ import annotations

from importlib.metadata import version
from pathlib import Path


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"
__version__: str = version("matrixctl")

HOME: str = str(Path.home())


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read and parse the configuration file with this module."""

from __future__ import annotations

import sys
import warnings

from copy import deepcopy
from logging import debug
from logging import error
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import toml

from matrixctl.errors import ConfigFileError


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


class TOML:

    """Use the TOML class to read and parse the configuration file(s)."""

    DEFAULT_PATHS: List[Path] = [
        Path("/etc/matrixctl/config"),
        Path.home() / ".config/matrixctl/config",
    ]
    __slots__ = ("__toml",)

    def __init__(self, path: Optional[Path] = None) -> None:
        debug("Loading Config file(s)")

        paths: List[Path] = []
        if path is None:
            paths += self.__class__.DEFAULT_PATHS

        self.__toml: Dict[str, Any] = self.__open(paths)

        self.__debug_output()

    @staticmethod
    def __open(paths: list[Path]) -> dict[str, Any]:
        """Open a TOML file and suppress warnings of the toml module.

        Parameters
        ----------
        paths : list of pathlib.Path
            A list of paths to check for the configuration files (toml) to
            combine into one config, which can be used by MatrixCtl.

        Returns
        -------
        toml : dict [str, any]
            The toml file structure represented as dict.

        """
        config_files: list[str] = [str(path) for path in paths]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                # toml.load is a MutableMapping
                # Only lists, tuples are not supported.
                return deepcopy(dict(toml.load(config_files)))
            except FileNotFoundError:
                error(
                    "To use this program you need to have a config file in"
                    '/etc/matrixctl/config" or in '
                    '"~/.config/matrixctl/config".'
                )
                sys.exit(1)
            except TypeError as e:
                raise ConfigFileError from e
            except toml.TomlDecodeError:
                error(
                    "Please check your config file. MatrixCtl was not able "
                    "to read it."
                )
                sys.exit(1)

    def __debug_output(self) -> None:
        """Create a debug output for the TOML file.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        for key in self.__toml:
            debug(f"[{key}]")

            for entry in self.__toml[key]:
                if entry == "Token":
                    length = len(self.__toml[key][entry])
                    debug(f"  ├─  {entry} := **HIDDEN (Length={length})**")
                else:
                    debug(f"  ├─  {entry} := {self.__toml[key][entry]}")
            debug("  ┴")

    # TODO: doctest + fixture
    def get(self, *keys: str) -> Any:
        """Get a value from a config entry safely.

        **Usage**

        Pass strings, describing the path in the ``self.__toml`` dictionary.
        Let's say, you are looking for the synapse path:

        Examples
        --------
        .. code-block:: python

           from matrixctl.handlers.toml import TOML

           toml: TOML = TOML()
           port: int = toml.get("SSH", "Port")
           print(port)
           # Output: 22

        Parameters
        ----------
        *keys : str
            A tuple of strings describing the values you are looking for.

        Returns
        -------
        answer : any
            The value of the entry you described.

        """
        toml_walker: Union[Dict[str, Any], Any] = self.__toml

        try:
            for key in keys:
                toml_walker = toml_walker.__getitem__(key)
        except KeyError:
            error(
                "Please check your config file. For this operation your "
                f'config file needs to have the entry "{keys[-1]}" '
                f'In the section "[{keys[0]}]".'
            )
            sys.exit(1)

        if not isinstance(toml_walker, dict):
            return toml_walker
        raise ConfigFileError(
            "The key you have asked for seems to be incorrect. "
            "Please make sure you ask for an single entry, "
            "not a entire section."
        )

    def __repr__(self) -> str:
        return repr(self.__toml)

    def __str__(self) -> str:
        return str(self.__toml)


#
# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``delroom`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import error

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_delroom(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl delroom`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by
        ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser(
        "delroom", help="Deletes an empty room from the database"
    )
    parser.add_argument(
        "RoomID",
        type=str,
        help="The Room-ID",
    )
    parser.set_defaults(func=delroom)


def delroom(arg: Namespace) -> int:
    """Delete an empty room from the database.

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    toml: TOML = TOML()
    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))
    api.method = "POST"
    api.url.path = "purge_room"
    api.url.api_version = "v1"

    try:
        api.request({"room_id": arg.RoomID}).json()
    except InternalResponseError as e:
        if "json" in dir(e.payload):
            try:
                if e.payload.json()["errcode"] in (
                    "M_NOT_FOUND",
                    "M_UNKNOWN",
                ):
                    error(f"{e.payload.json()['error']}")

                    return 1
            except KeyError:
                pass  # log the fallback error

        error("Could not delete room")

        return 1

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module as entrypoint for the application."""

from __future__ import annotations

import argparse

from argparse import _SubParsersAction
from logging import debug
from logging import warning
from typing import Callable
from typing import List

import coloredlogs

from matrixctl import __version__
from matrixctl.adduser import subparser_adduser
from matrixctl.adduser_jitsi import subparser_adduser_jitsi
from matrixctl.check import subparser_check
from matrixctl.delroom import subparser_delroom
from matrixctl.deluser import subparser_deluser
from matrixctl.deluser_jitsi import subparser_deluser_jitsi
from matrixctl.deploy import subparser_deploy
from matrixctl.maintenance import subparser_maintenance
from matrixctl.rooms import subparser_rooms
from matrixctl.server_notice import subparser_server_notice
from matrixctl.start import subparser_restart
from matrixctl.start import subparser_start
from matrixctl.update import subparser_update
from matrixctl.upload import subparser_upload
from matrixctl.user import subparser_user
from matrixctl.users import subparser_users
from matrixctl.version import subparser_version


# Subparsers


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"

# API: https://github.com/matrix-org/synapse/blob/master/docs/admin_api/
#              user_admin_api.rst


def setup_parser() -> argparse.ArgumentParser:
    """Use this class to initialize the parser and the subparsers.

    Parameters
    ----------
    None

    Returns
    -------
    parser : argparse.ArgumentParser
        The parser object, which can be used to parse the arguments.

    """
    parser = argparse.ArgumentParser()

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enables debugging mode."
    )
    subparsers: _SubParsersAction = parser.add_subparsers()

    # Subparsers
    subparsers_tuple: List[Callable[[_SubParsersAction], None]] = [
        subparser_adduser,
        subparser_deluser,
        subparser_adduser_jitsi,
        subparser_deluser_jitsi,
        subparser_user,
        subparser_users,
        subparser_rooms,
        subparser_delroom,
        subparser_update,
        subparser_upload,
        subparser_deploy,
        subparser_server_notice,
        subparser_start,
        subparser_restart,  # alias for start
        subparser_maintenance,
        subparser_check,
        subparser_version,
    ]

    for subparser in subparsers_tuple:
        subparser(subparsers)

    return parser


def setup_logging(debug_mode: bool) -> None:
    """Use this function to setup logging for the application.

    Parameters
    ----------
    debug_mode : bool
        ``True`` sets the log level to ``DEBUG``, ``False`` sets the log level
        to ``INFO``.

    Returns
    -------
    None

    """
    coloredlogs.DEFAULT_LOG_FORMAT = (
        "%(asctime)s [%(module)s:%(lineno)d (%(funcName)s)] - %(levelname)s "
        "- %(message)s"
    )
    coloredlogs.DEFAULT_LOG_LEVEL = 0 if debug_mode else 21
    coloredlogs.install()


def main() -> int:
    """Use the ``main`` function as entrypoint to run the application.

    Parameters
    ----------
    None

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    parser = setup_parser()

    args: argparse.Namespace = parser.parse_args()

    setup_logging(args.debug)

    debug(f"{args=}")

    if args.debug:
        debug("Disabing help on AttributeError")
        warning(
            "In debugging mode help is disabled! If you don't use any "
            "attibutes, the program will throw a AttributeError like: "
            "\"AttributeError: 'Namespace' object has no attribute 'func\".'"
            " This is perfectly normal and not a bug. If you want the help "
            'in debug mode, use the "--help" attribute.'
        )

        return int(args.func(args))

    try:
        return int(args.func(args))
    except AttributeError:
        parser.print_help()

        return 1


if __name__ == "__main__":
    import sys

    sys.exit(main())

# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Run and evaluate commands on the host machine of your synapse server."""

from __future__ import annotations

from getpass import getuser
from logging import debug
from types import TracebackType
from typing import NamedTuple
from typing import Optional
from typing import Type

from paramiko import SSHClient
from paramiko.channel import ChannelFile


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


class SSHResponse(NamedTuple):

    """Store the response of a SSH command as response."""

    stdin: Optional[str]
    stdout: Optional[str]
    stderr: Optional[str]


class SSH:

    """Run and evaluate commands on the host machine of your synapse server."""

    __slots__ = ("address", "__client", "user", "port")

    def __init__(
        self, address: str, user: Optional[str] = None, port: int = 22
    ) -> None:
        self.address: str = address
        self.port: int = port
        self.user: str = getuser() if user is None else user
        self.__client: SSHClient = SSHClient()
        self.__client.load_system_host_keys()
        self.__connect()

    def __connect(self) -> None:
        """Connect to the SSH server.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.__client.connect(self.address, self.port, self.user)
        debug("SSH connected")

    def __disconnect(self) -> None:
        """Disconnect from the SSH server.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.__client.close()
        debug("SSH disconnected")

    @staticmethod
    def __str_from(f: ChannelFile) -> str | None:
        """Convert a ChannelFile to str.

        Parameters
        ----------
        f : paramiko.channel.ChannelFile
            ``stdin``, ``stdout`` or ``stderr`` as ChannelFile.

        Returns
        -------
        response_str : str, optional
            ``stdin``, ``stdout`` or ``stderr`` as str.

        """
        try:
            return str(f.read().decode("utf-8").strip())
        except OSError:
            return None

    def run_cmd(self, cmd: str) -> SSHResponse:
        """Run a command on the host machine and receive a response.

        Parameters
        ----------
        cmd : str
            The command to run.

        Returns
        -------
        response : matrixctl.handlers.ssh.SSHResponse
            Receive ``stdin``, ``stdout`` and ``stderr`` as response.

        """
        debug(f'SSH Command: "{cmd}"')

        response: SSHResponse = SSHResponse(
            # skipcq: BAN-B601
            *[self.__str_from(s) for s in self.__client.exec_command(cmd)]
        )

        debug(f'SSH Response: "{response}"')

        return response

    def __enter__(self) -> SSH:
        """Connect to the SSH server with the "with" statement.

        Parameters
        ----------
        None

        Returns
        -------
        ssh_instance : matrixctl.handlers.ssh.SSH
            The object itself.

        """

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Close the SSH connection.

        Parameters
        ----------
        exc_type : types.Type [BaseException], optional
            (Unused)
        exc_val : BaseException, optional
            (Unused)
        exc_tb : types.TracebackType, optional
            (Unused)

        Returns
        -------
        None

        """
        debug(f"SSH __exit__: {exc_type=}, {exc_val=}, {exc_tb=}")
        self.__disconnect()

    def __del__(self) -> None:
        """Close the connection to the SSH.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.__disconnect()


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``users`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import fatal
from typing import List
from typing import Tuple

from tabulate import tabulate

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML
from .print_helpers import human_readable_bool
from .typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_users(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl users`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser("users", help="Lists users")
    parser.add_argument(
        "-a", "--all", action="store_true", help="Shows all users"
    )
    parser.add_argument(
        "-g", "--with-guests", action="store_true", help="Shows guests"
    )
    parser.add_argument(
        "-d",
        "--with-deactivated",
        action="store_true",
        help="Shows deactivated accounts",
    )
    parser.set_defaults(func=users)


def users(arg: Namespace) -> int:
    """Print a table of the matrix users.

    This function generates and prints a table of matrix user accounts.
    The table can be modified.

    - If you want guests in the table use the ``--with-guests`` switch.
    - If you want deactivated user in the table use the ``--with-deactivated``
      switch.

    Examples
    --------
    .. code-block:: console

       $ matrixctl users
       +----------------+---------------+------------+------------+
       | Name           | Deactivated   | Is Admin   | Is Guest   |
       |----------------+---------------+------------+------------|
       | dunder_mifflin | False         | True       | False      |
       | dwight         | False         | True       | False      |
       | pam            | False         | False      | False      |
       | jim            | False         | False      | False      |
       | creed          | False         | False      | False      |
       | stanley        | False         | False      | False      |
       | kevin          | False         | False      | False      |
       | angela         | False         | False      | False      |
       | phyllis        | False         | False      | False      |
       | tobi           | False         | False      | False      |
       | michael        | False         | True       | False      |
       | andy           | False         | False      | False      |
       +----------------+---------------+------------+------------+

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``.

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    toml: TOML = TOML()
    len_domain = len(toml.get("API", "Domain")) + 1  # 1 for :
    from_user: int = 0
    users_list: List[JsonDict] = []

    # ToDo: API bool
    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))
    api.url.path = "users"
    api.params = {"guests": "true" if arg.with_guests or arg.all else "false"}
    api.params = {
        "deactivated": "true" if arg.with_deactivated or arg.all else "false"
    }

    while True:

        api.params = {"from": from_user}  # from must be in the loop
        try:
            lst: JsonDict = api.request().json()
        except InternalResponseError:
            fatal("Could not get the user table.")

            return 1

        users_list += lst["users"]

        try:
            from_user = lst["next_token"]
        except KeyError:
            break

    user_list: List[Tuple[str, str, str, str]] = []

    for user in users_list:
        name = user["name"][1:-len_domain]
        deactivated: str = human_readable_bool(user["deactivated"])
        admin: str = human_readable_bool(user["admin"])
        guest: str = human_readable_bool(user["is_guest"])

        user_list.append(
            (
                name,
                deactivated,
                admin,
                guest,
            )
        )
    print(
        tabulate(
            user_list,
            headers=("Name", "Deactivated", "Is Admin", "Is Guest"),
            tablefmt="psql",
        )
    )

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``rooms`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import fatal
from typing import List

from tabulate import tabulate

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML
from .typing import JsonDict


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_rooms(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl rooms`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser("rooms", help="List rooms")
    parser.add_argument(
        "-s",
        "--order-by-size",
        action="store_true",
        help="Order the rooms by size instead of alphabetical",
    )
    parser.add_argument(
        "-r", "--reverse", action="store_true", help="Reverse the order"
    )
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        default=0,
        help="The number of rooms to show",
    )
    parser.add_argument(
        "filter",
        type=str,
        nargs="?",
        default=None,
        help="Search for rooms with a specific search term (Case-Sensitive)",
    )
    parser.set_defaults(func=rooms)


def rooms(arg: Namespace) -> int:
    """Generate a table of the matrix rooms.

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``.

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    toml: TOML = TOML()
    from_room: int = 0
    rooms_list: List[JsonDict] = []

    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))
    api.url.path = "rooms"
    api.url.api_version = "v1"

    if arg.number > 0:
        api.params = {"limit": arg.number}

    if arg.filter:
        api.params = {"search_term": arg.filter}

    if arg.reverse:
        api.params = {"dir": "b"}

    if arg.order_by_size:
        api.params = {"order_by": "size"}

    while True:

        api.params = {"from": from_room}  # from must be in the loop
        try:
            lst: JsonDict = api.request().json()
        except InternalResponseError:
            fatal("Could not get the room table.")

            return 1

        rooms_list += lst["rooms"]
        try:
            from_room = lst["next_token"]
        except KeyError:
            break
    print_rooms_table(rooms_list)

    return 0


def print_rooms_table(rooms_list: list[JsonDict]) -> None:
    """Use this function as helper to pint the room table.

    Parameters
    ----------
    rooms_list : list of matrixctl.tping.JsonDict
        A list of rooms from the API.

    Returns
    -------
    None

    """

    room_list: list[tuple[str, int, str, str]] = []

    for room in rooms_list:
        name = room["name"]
        members: int = room["joined_members"]
        alias: str = room["canonical_alias"]
        room_id: str = room["room_id"]

        room_list.append(
            (
                name,
                members,
                alias,
                room_id,
            )
        )
    print(
        tabulate(
            room_list,
            headers=("Name", "Members", "Alias", "Room ID"),
            tablefmt="psql",
        )
    )


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Update and manage the synapse playbook repository with this module."""

from __future__ import annotations

import datetime
import sys

from logging import critical
from logging import debug
from logging import error
from logging import info
from pathlib import Path
from shutil import get_terminal_size
from textwrap import TextWrapper
from typing import List
from typing import Union

import git

from tabulate import tabulate


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


class Git:

    """Update and manage a repository."""

    def __init__(self, path: Union[Path, str]) -> None:
        self.path: Path = Path(path)
        self.repo = git.Repo(self.path)

        if self.repo.bare:
            critical(
                "Please make sure you entered the correct repository "
                "in [SYNAPSE] -> Playbook."
            )
            sys.exit(1)

        self.git = git.cmd.Git(self.path)
        self.heads = self.repo.heads
        self.master = self.heads.master

    @property
    def datetime_last_pulled_commit(self) -> datetime.datetime:
        """Get the datetime the commit was pulled last from git.

        This is used to determine which messages will be produced in the table.

        Parameters
        ----------
        None

        Returns
        -------
        datetime : datetime.datetime
            The datetime object.

        """
        log = self.master.log()

        return datetime.datetime.fromtimestamp(log[-1].time[0])

    def log(self, since: datetime.datetime | None = None) -> None:
        """Print a table of date, user and commit message since the last pull.

        Parameters
        ----------
        since : datetime.datetime, optional, default=None
            The datetime the last commit was puled.

        Returns
        -------
        None

        """
        cmd = ["--pretty=%as\t%an\t%s"]

        if since:
            cmd.append(f"--since={str(since)}")

        terminal_size_x, _ = get_terminal_size()
        debug(f"Terminal width = {terminal_size_x}")

        ######################################################################
        #                          Terminal width                            #
        #                          ^^^^^^^^^^^^^^                            #
        #                                                                    #
        # |<-------------------------------(121)-------------------------->| #
        # |<--------------------------(119)------------------------->|     | #
        # |<----------------------(34)--------------------->|        |     | #
        # |<------------------(32)----------------->|       |        |     | #
        # |<---------------(30)-------------->|     |       |        |     | #
        # |<-------------16------------>|     |     |       |        |     | #
        # |<---------14--------->|      |     |     |       |        |     | #
        # |<----12----->|        |      |     |     |       |        |     | #
        # |<--3-->|     |        |      |     |     |       |        |     | #
        # |_______d a t e________|______u s e r_____|_______commit_msg_____| #
        #                               |<--->|            |<------>|        #
        #                                 15                 x - 35          #
        #                                                                    #
        ######################################################################

        wrapper_user = TextWrapper(
            width=15, drop_whitespace=True, break_long_words=True
        )
        wrapper_comment = TextWrapper(
            width=terminal_size_x - 35,
            drop_whitespace=True,
            break_long_words=True,
        )

        log: List[List[str]] = [
            line.split("\t") for line in self.git.log(cmd).split("\n")
        ]

        if not log[0][0]:  # Nothing new
            info("Everything is up-to-date.")

            return

        for line in log:
            line[1] = wrapper_user.fill(text=line[1])
            line[2] = wrapper_comment.fill(text=line[2])

        print(
            tabulate(
                log,
                headers=("Date", "User", "Commit Message"),
                tablefmt="psql",
            )
        )

    def pull(self) -> None:
        """Git pull the latest commits from GH.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # Get the last pulled datetime
        since = self.datetime_last_pulled_commit

        try:
            self.git.pull()
        except git.GitCommandError:
            error(
                "MatrixCtl was not able to connect to the synapse playbook "
                "on GitHub. Are you connected to the internet?"
            )
            sys.exit(1)

        self.log(since)


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use this module to add the ``deluser`` subcommand to ``matrixctl``."""

from __future__ import annotations

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _SubParsersAction as SubParsersAction
from logging import error

from .errors import InternalResponseError
from .handlers.api import API
from .handlers.toml import TOML


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


def subparser_deluser(subparsers: SubParsersAction) -> None:
    """Create a subparser for the ``matrixctl deluser`` command.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The object which is returned by
        ``parser.add_subparsers()``.

    Returns
    -------
    None

    """
    parser: ArgumentParser = subparsers.add_parser(
        "deluser", help="Deletes a user"
    )
    parser.add_argument("user", help="The username to delete")
    parser.set_defaults(func=deluser)


def deluser(arg: Namespace) -> int:
    """Delete a user from the the matrix instance.

    Parameters
    ----------
    arg : argparse.Namespace
        The ``Namespace`` object of argparse's ``parse_args()``

    Returns
    -------
    err_code : int
        Non-zero value indicates error code, or zero on success.

    """
    toml: TOML = TOML()
    api: API = API(toml.get("API", "Domain"), toml.get("API", "Token"))
    try:
        api.url.path = f"deactivate/@{arg.user}:{toml.get('API','Domain')}"
        api.url.api_version = "v1"
        api.method = "POST"
        api.request({"erase": True})
    except InternalResponseError:
        error("The user was not deleted.")

    return 0


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use the functions of this module as helpers for passwords."""

from __future__ import annotations

import getpass
import secrets
import string

from logging import info
from typing import Optional


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


SPECIAL = "!§$%&/()=?.,;:_-#+*~{}[]°^@<>|\\"
ALPHABET = string.ascii_letters + string.digits + SPECIAL


def ask_password() -> Optional[str]:
    """Ask the user to create a password.

    The user will be asked twice, after
    that, the function compares the two entered passwords. If they are the
    same, the function will return the password.

    Notes
    -----
    If the user presses enter twice, without entering a password,
    the function will return ``None``. This is by-design.

    Parameters
    ----------
    None

    Returns
    -------
    password : str, optional
        The user entered password or ``None``, if the password does not match
        twice.

    """
    # ToDo: Check password (regex)
    passwd: str = "a"
    passwd2: str = "b"  # Something invalid: a==b is not True

    while passwd != passwd2:
        passwd = getpass.getpass()
        passwd2 = getpass.getpass("Password (again): ")

    return passwd if passwd == passwd2 else None


def gen_password(
    length: int = 16,
    min_digits: int = 3,
    min_special: int = 1,
    max_special: int = 3,
) -> str:
    """Generate a password.

    It uses lower- and uppercase
    characters, digits and some special characters, you can find in the
    ``SPECIAL`` variable in this file.

    It uses at least:

    - ``min_digits`` digits
    - ``min_special`` special characters

    It uses maximal:

    - ``max_special`` special characters

    The rest are lower- and uppercase characters.

    Parameters
    ----------
    length : int, defautl=16
        The desired length of the password to generate.
    min_digits : int, default=3
        The desired minimum of digits of the password.
    min_special : int, default=1
        The desired minimum of special characters of the password.
    max_special : int, default=3
        The desired maximum of special characters of the password.

    Returns
    -------
    password : str
        The generated password.

    """

    while True:
        password = "".join(secrets.choice(ALPHABET) for i in range(length))

        # pylint: disable=chained-comparison

        if (
            any(c.islower() for c in password)
            and any(c.isupper() for c in password)
            and sum(c.isdigit() for c in password) >= min_digits
            and min_special
            <= sum(c in SPECIAL for c in password)
            <= max_special
        ):
            return password


def ask_question(question: str = "Is everything correct?") -> bool:
    """Asks the user a simple yes/no a question.

    Notes
    -----
    - The user entered value is case-insensitive.
    - If the user answered with an invalid answer (not ``y`` / ``j`` / ``n``)
      the function asks again.

    Parameters
    ----------
    question : str
        The yes/no question the user should be asked

    Returns
    -------
    answer : bool
        ``True`` if the answer was ``y`` / ``j``, or
        ``False`` if the answer was ``n``

    """
    question += " [y/n]"

    while (answer := input(question).lower()) not in (
        "y",
        "j",
        "n",
    ):
        info("User entered [y/n] pattern did not match")

    return answer in ("y", "j")


# vim: set ft=python :



#!/usr/bin/env python
# matrixctl
# Copyright (c) 2020  Michael Sasser <Michael@MichaelSasser.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Run a ansible playbook with this module."""

from __future__ import annotations

from logging import debug
from pathlib import Path

from ansible_runner.interface import Runner
from ansible_runner.runner_config import RunnerConfig


__author__: str = "Michael Sasser"
__email__: str = "Michael@MichaelSasser.org"


# ToDo: Make async to get debug output while running
def ansible_run(
    playbook: Path,
    tags: str | None = None,
    extra_vars: dict[str, str] | None = None,
) -> None:
    """Run an ansible playbook.

    Parameters
    ----------
    playbook : pathlib.Path
        The path to the ansible Playbook
    tags : str, optional
        The tags to use
    extra_vars : dict [str, str], optional
        The extra_vars to use.

    Returns
    -------
    None

    """
    runner_config: RunnerConfig = RunnerConfig(
        private_data_dir="/tmp",
        playbook=playbook,
        tags=tags,
        extravars=extra_vars,
    )
    runner_config.prepare()

    runner: Runner = Runner(config=runner_config)
    runner.run()

    # debug output
    debug("Runner status")
    debug(f"{runner.status}: {runner.rc}")
    for host_event in runner.events:
        debug(host_event["event"])
    debug(f"Final status: {runner.stats}")


# vim: set ft=python :
